@@ -241,30 +241,43 @@ void view_load(int view, char *arg)
 	}
 }
 
-static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filenamep)
+static void do_save(for_each_ti_cb for_each_ti, const char *arg, char **filenamep, save_ti_cb save_ti)
 {
 	char *filename = *filenamep;
 
 	if (arg) {
-		free(filename);
-		filename = xstrdup(arg);
-		*filenamep = filename;
+		if (strcmp(arg, "-") == 0) {
+			filename = (char *) arg;
+		} else {
+			free(filename);
+			filename = xstrdup(arg);
+			*filenamep = filename;
+		}
+	} else if (!filename) {
+		error_msg("need a file as argument, no default stored yet");
+		return;
 	}
 
 	editable_lock();
-	if (cmus_save(for_each_ti, filename) == -1)
+	if (save_ti(for_each_ti, filename) == -1)
 		error_msg("saving '%s': %s", filename, strerror(errno));
 	editable_unlock();
 }
 
-void view_save(int view, char *arg)
+void view_save(int view, char *arg, int to_stdout, int filtered, int extended)
 {
-	if (arg) {
-		char *tmp;
+	char **dest;
+	save_ti_cb     save_ti         = extended ? cmus_save_ext         : cmus_save;
+	for_each_ti_cb lib_for_each_ti = filtered ? lib_for_each_filtered : lib_for_each;
 
-		tmp = expand_filename(arg);
-		arg = path_absolute(tmp);
-		free(tmp);
+	if (arg) {
+		if (to_stdout) {
+			arg = xstrdup(arg);
+		} else {
+			char *tmp = expand_filename(arg);
+			arg = path_absolute(tmp);
+			free(tmp);
+		}
 	}
 
 	switch (view) {
@@ -272,12 +285,20 @@ void view_save(int view, char *arg)
 	case SORTED_VIEW:
 		if (worker_has_job(JOB_TYPE_LIB))
 			goto worker_running;
-		do_save(lib_for_each, arg, &lib_filename);
+		dest = extended ? &lib_ext_filename : &lib_filename;
+		do_save(lib_for_each_ti, arg, dest, save_ti);
 		break;
 	case PLAYLIST_VIEW:
 		if (worker_has_job(JOB_TYPE_PL))
 			goto worker_running;
-		do_save(pl_for_each, arg, &pl_filename);
+		dest = extended ? &pl_ext_filename : &pl_filename;
+		do_save(pl_for_each, arg, &pl_filename, save_ti);
+		break;
+	case QUEUE_VIEW:
+		if (worker_has_job(JOB_TYPE_QUEUE))
+			goto worker_running;
+		dest = extended ? &play_queue_ext_filename : &play_queue_filename;
+		do_save(play_queue_for_each, arg, dest, save_ti);
 		break;
 	default:
 		info_msg(":save only works in views 1 & 2 (library) and 3 (playlist)");
@@ -291,8 +312,8 @@ worker_running:
 
 /* }}} */
 
-/* only returns the last flag which is enough for it's callers */
-static int parse_flags(const char **strp, const char *flags)
+/* if only_last != 0, only return the last flag */
+static int do_parse_flags(const char **strp, const char *flags, int only_last)
 {
 	const char *str = *strp;
 	int flag = 0;
@@ -300,7 +321,7 @@ static int parse_flags(const char **strp, const char *flags)
 	if (str == NULL)
 		return flag;
 
-	while (*str) {
+	while (*str && (only_last || !flag)) {
 		if (*str != '-')
 			break;
 
@@ -337,10 +358,47 @@ static int parse_flags(const char **strp, const char *flags)
 	return flag;
 }
 
+static int parse_flags(const char **strp, const char *flags)
+{
+	return do_parse_flags(strp, flags, 1);
+}
+
+static int parse_one_flag(const char **strp, const char *flags)
+{
+	return do_parse_flags(strp, flags, 0);
+}
+
+/* is str == "...-", but not "...-- -" ? copied from do_parse_flags() */
+static int is_stdout_filename(const char *str)
+{
+	if (!str)
+		return 0;
+
+	while (*str) {
+		if (*str != '-')
+			return 0;
+		// "-"
+		if (str[1] == 0)
+			return 1;
+		// "--" or "-- "
+		if (str[1] == '-' && (str[2] == 0 || str[2] == ' '))
+			return 0;
+		// not "-?" or "-? "
+		if (str[2] && str[2] != ' ')
+			return 0;
+		str += 2;
+		while (*str == ' ')
+			str++;
+	}
+
+	return 0;
+}
+
 static int flag_to_view(int flag)
 {
 	switch (flag) {
 	case 'l':
+	case 'L':
 		return TREE_VIEW;
 	case 'p':
 		return PLAYLIST_VIEW;
@@ -393,11 +451,20 @@ static void cmd_load(char *arg)
 
 static void cmd_save(char *arg)
 {
-	int flag = parse_flags((const char **)&arg, "lp");
+	int to_stdout = is_stdout_filename(arg);
+	int flag = 0, f, extended = 0;
+
+	do {
+		f = parse_one_flag((const char **)&arg, "eLlpq");
+		if (f == 'e')
+			extended = 1;
+		else if (f)
+			flag = f;
+	} while (f > 0);
 
 	if (flag == -1)
 		return;
-	view_save(flag_to_view(flag), arg);
+	view_save(flag_to_view(flag), arg, to_stdout, flag == 'L', extended);
 }
 
 static void cmd_set(char *arg)
@@ -508,7 +575,7 @@ inside:
 	}
 
 	if (!*arg) {
-		player_seek(seek, relative);
+		player_seek(seek, relative, 1);
 		return;
 	}
 err:
@@ -522,6 +589,13 @@ static void cmd_factivate(char *arg)
 	editable_unlock();
 }
 
+static void cmd_live_filter(char *arg)
+{
+	editable_lock();
+	filters_set_live(arg);
+	editable_unlock();
+}
+
 static void cmd_filter(char *arg)
 {
 	editable_lock();
@@ -1022,7 +1096,7 @@ static void cmd_run(char *arg)
 	if (run) {
 		int status;
 
-		if (spawn(argv, &status)) {
+		if (spawn(argv, &status, 1)) {
 			error_msg("executing %s: %s", argv[0], strerror(errno));
 		} else {
 			if (WIFEXITED(status)) {
@@ -1056,6 +1130,14 @@ static void cmd_run(char *arg)
 	free(sel_tis);
 }
 
+static void cmd_shell(char *arg)
+{
+	const char * const argv[] = { "sh", "-c", arg, NULL };
+
+	if (spawn((char **) argv, NULL, 0))
+		error_msg("executing '%s': %s", arg, strerror(errno));
+}
+
 static int get_one_ti(void *data, struct track_info *ti)
 {
 	struct track_info **sel_ti = data;
@@ -1203,10 +1285,16 @@ static void cmd_vol(char *arg)
 		r = calc_vol(r, soft_vol_r, 100, rf);
 		player_set_soft_volume(l, r);
 	} else {
+		int rc;
 		mixer_read_volume();
 		l = calc_vol(l, volume_l, volume_max, lf);
 		r = calc_vol(r, volume_r, volume_max, rf);
-		mixer_set_volume(l, r);
+		rc = mixer_set_volume(l, r);
+		if (rc != OP_ERROR_SUCCESS) {
+			char *msg = op_get_error_msg(rc, "can't change volume");
+			error_msg("%s", msg);
+			free(msg);
+		}
 		mixer_read_volume();
 	}
 	update_statusline();
@@ -1309,12 +1397,17 @@ static for_each_sel_ti_cb view_for_each_sel[4] = {
 	pq_for_each_sel
 };
 
+/* wrapper for add_ti_cb, (void *) can't store function pointers */
+struct wrapper_cb_data {
+	add_ti_cb cb;
+};
+
 /* wrapper for void lib_add_track(struct track_info *) etc. */
 static int wrapper_cb(void *data, struct track_info *ti)
 {
-	add_ti_cb add = data;
+	struct wrapper_cb_data *add = data;
 
-	add(ti);
+	add->cb(ti);
 	return 0;
 }
 
@@ -1342,8 +1435,9 @@ static void cmd_win_add_l(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { lib_add_track };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, lib_add_track, 0);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(lib_add_track, JOB_TYPE_LIB);
@@ -1357,8 +1451,9 @@ static void cmd_win_add_p(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { pl_add_track };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, pl_add_track, 0);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(pl_add_track, JOB_TYPE_PL);
@@ -1371,8 +1466,9 @@ static void cmd_win_add_Q(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { play_queue_prepend };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, play_queue_prepend, 1);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 1);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(play_queue_prepend, JOB_TYPE_QUEUE);
@@ -1385,8 +1481,9 @@ static void cmd_win_add_q(char *arg)
 		return;
 
 	if (cur_view <= QUEUE_VIEW) {
+		struct wrapper_cb_data add = { play_queue_append };
 		editable_lock();
-		view_for_each_sel[cur_view](wrapper_cb, play_queue_append, 0);
+		view_for_each_sel[cur_view](wrapper_cb, &add, 0);
 		editable_unlock();
 	} else if (cur_view == BROWSER_VIEW) {
 		add_from_browser(play_queue_append, JOB_TYPE_QUEUE);
@@ -1734,11 +1831,11 @@ found:
 static int count_albums(void)
 {
 	struct artist *artist;
-	struct list_head *item;
+	struct rb_node *tmp1, *tmp2;
 	int count = 0;
 
-	list_for_each_entry(artist, &lib_artist_head, node) {
-		list_for_each(item, &artist->album_head)
+	rb_for_each_entry(artist, tmp1, &lib_artist_root, tree_node) {
+		rb_for_each(tmp2, &artist->album_root)
 			count++;
 	}
 	return count;
@@ -1774,18 +1871,18 @@ static void cmd_lqueue(char *arg)
 		goto unlock;
 
 	r = rand_array(count, nmax);
-	album = to_album(to_artist(lib_artist_head.next)->album_head.next);
+	album = to_album(rb_first(&to_artist(rb_first(&lib_artist_root))->album_root));
 	pos = 0;
 	for (i = 0; i < count; i++) {
 		struct album_list *a;
 
 		while (pos < r[i]) {
 			struct artist *artist = album->artist;
-			if (album->node.next == &artist->album_head) {
-				artist = to_artist(artist->node.next);
-				album = to_album(artist->album_head.next);
+			if (!rb_next(&album->tree_node)) {
+				artist = to_artist(rb_next(&artist->tree_node));
+				album = to_album(rb_first(&artist->album_root));
 			} else {
-				album = to_album(album->node.next);
+				album = to_album(rb_next(&album->tree_node));
 			}
 			pos++;
 		}
@@ -1800,8 +1897,9 @@ static void cmd_lqueue(char *arg)
 		struct list_head *next = item->next;
 		struct album_list *a = container_of(item, struct album_list, node);
 		struct tree_track *t;
-
-		list_for_each_entry(t, &a->album->track_head, node)
+		struct rb_node *tmp;
+ 
+		rb_for_each_entry(t, tmp, &album->track_root, tree_node)
 			editable_add(&pq_editable, simple_track_new(tree_track_info(t)));
 		free(a);
 		item = next;
@@ -1955,7 +2053,7 @@ static void expand_key_context(const char *str, const char *force)
 	int pos, i, len = strlen(str);
 	char **tails;
 
-	tails = xnew(char *, NR_CTXS + 1);
+	tails = xnew(char *, NR_CTXS);
 	pos = 0;
 	for (i = 0; key_context_names[i]; i++) {
 		int cmp = strncmp(str, key_context_names[i], len);
@@ -1975,10 +2073,10 @@ static void expand_key_context(const char *str, const char *force)
 		free(tails[0]);
 		tails[0] = tmp;
 	}
-	tails[pos] = NULL;
 	snprintf(expbuf, sizeof(expbuf), "%s%s", force, str);
 	tabexp.head = xstrdup(expbuf);
 	tabexp.tails = tails;
+	tabexp.count = pos;
 }
 
 static int get_context(const char *str, int len)
@@ -2071,9 +2169,9 @@ static void expand_bind_args(const char *str)
 
 		snprintf(expbuf, sizeof(expbuf), "%s%s %s", force, key_context_names[c], ks);
 
-		ptr_array_plug(&array);
 		tabexp.head = xstrdup(expbuf);
 		tabexp.tails = array.ptrs;
+		tabexp.count = array.count;
 		return;
 	}
 
@@ -2164,9 +2262,9 @@ static void expand_unbind_args(const char *str)
 
 	snprintf(expbuf, sizeof(expbuf), "%s %s", key_context_names[c], ks);
 
-	ptr_array_plug(&array);
 	tabexp.head = xstrdup(expbuf);
 	tabexp.tails = array.ptrs;
+	tabexp.count = array.count;
 }
 
 static void expand_factivate(const char *str)
@@ -2194,9 +2292,9 @@ static void expand_factivate(const char *str)
 	if (!array.count)
 		return;
 
-	ptr_array_plug(&array);
 	tabexp.head = xstrdup(str);
 	tabexp.tails = array.ptrs;
+	tabexp.count = array.count;
 }
 
 static void expand_options(const char *str)
@@ -2215,15 +2313,15 @@ static void expand_options(const char *str)
 			if (strcmp(var, opt->name) == 0) {
 				char buf[OPTION_MAX_SIZE];
 
-				tails = xnew(char *, 2);
+				tails = xnew(char *, 1);
 
 				buf[0] = 0;
 				opt->get(opt->id, buf);
 				tails[0] = xstrdup(buf);
-				tails[1] = NULL;
 
 				tabexp.head = xstrdup(str);
 				tabexp.tails = tails;
+				tabexp.count = 1;
 				free(var);
 				return;
 			}
@@ -2233,7 +2331,7 @@ static void expand_options(const char *str)
 		/* expand variable */
 		int pos;
 
-		tails = xnew(char *, nr_options + 1);
+		tails = xnew(char *, nr_options);
 		pos = 0;
 		list_for_each_entry(opt, &option_head, node) {
 			if (strncmp(str, opt->name, len) == 0)
@@ -2248,9 +2346,9 @@ static void expand_options(const char *str)
 				tails[0] = tmp;
 			}
 
-			tails[pos] = NULL;
 			tabexp.head = xstrdup(str);
 			tabexp.tails = tails;
+			tabexp.count = pos;
 		} else {
 			free(tails);
 		}
@@ -2263,7 +2361,7 @@ static void expand_toptions(const char *str)
 	int len, pos;
 	char **tails;
 
-	tails = xnew(char *, nr_options + 1);
+	tails = xnew(char *, nr_options);
 	len = strlen(str);
 	pos = 0;
 	list_for_each_entry(opt, &option_head, node) {
@@ -2273,9 +2371,9 @@ static void expand_toptions(const char *str)
 			tails[pos++] = xstrdup(opt->name + len);
 	}
 	if (pos > 0) {
-		tails[pos] = NULL;
 		tabexp.head = xstrdup(str);
 		tabexp.tails = tails;
+		tabexp.count = pos;
 	} else {
 		free(tails);
 	}
@@ -2319,9 +2417,9 @@ static void expand_colorscheme(const char *str)
 	if (array.count) {
 		ptr_array_sort(&array, strptrcmp);
 
-		ptr_array_plug(&array);
 		tabexp.head = xstrdup(str);
 		tabexp.tails = array.ptrs;
+		tabexp.count = array.count;
 	}
 }
 
@@ -2343,6 +2441,7 @@ struct command commands[] = {
 	{ "fset",		cmd_fset,	1, 1, NULL,		  0, 0 },
 	{ "help",		cmd_help,	0, 0, NULL,		  0, 0 },
 	{ "invert",		cmd_invert,	0, 0, NULL,		  0, 0 },
+	{ "live-filter",	cmd_live_filter,0, 1, NULL,		  0, CMD_LIVE },
 	{ "load",		cmd_load,	1, 1, expand_load_save,	  0, 0 },
 	{ "lqueue",		cmd_lqueue,	0, 1, NULL,		  0, 0 },
 	{ "mark",		cmd_mark,	0, 1, NULL,		  0, 0 },
@@ -2361,6 +2460,7 @@ struct command commands[] = {
 	{ "search-prev",	cmd_search_prev,0, 0, NULL,		  0, 0 },
 	{ "seek",		cmd_seek,	1, 1, NULL,		  0, 0 },
 	{ "set",		cmd_set,	1, 1, expand_options,	  0, 0 },
+	{ "shell",		cmd_shell,	1,-1, NULL,		  0, CMD_UNSAFE },
 	{ "showbind",		cmd_showbind,	1, 1, expand_unbind_args, 0, 0 },
 	{ "shuffle",		cmd_reshuffle,	0, 0, NULL,		  0, 0 },
 	{ "source",		cmd_source,	1, 1, expand_files,	  0, CMD_UNSAFE },
@@ -2371,6 +2471,7 @@ struct command commands[] = {
 	{ "update-cache",	cmd_update_cache,0, 0, NULL,		  0, 0 },
 	{ "view",		cmd_view,	1, 1, NULL,		  0, 0 },
 	{ "vol",		cmd_vol,	1, 2, NULL,		  0, 0 },
+	{ "w",			cmd_save,	0, 1, expand_load_save,	  0, CMD_UNSAFE },
 	{ "win-activate",	cmd_win_activate,0, 0, NULL,		  0, 0 },
 	{ "win-add-l",		cmd_win_add_l,	0, 0, NULL,		  0, 0 },
 	{ "win-add-p",		cmd_win_add_p,	0, 0, NULL,		  0, 0 },
@@ -2389,6 +2490,7 @@ struct command commands[] = {
 	{ "win-top",		cmd_win_top,	0, 0, NULL,		  0, 0 },
 	{ "win-up",		cmd_win_up,	0, 0, NULL,		  0, 0 },
 	{ "win-update",		cmd_win_update,	0, 0, NULL,		  0, 0 },
+	{ "wq",			cmd_quit,	0, 1, NULL,		  0, 0 },
 	{ NULL,			NULL,		0, 0, 0,		  0, 0 }
 };
 
@@ -2399,7 +2501,7 @@ static void expand_commands(const char *str)
 	char **tails;
 
 	/* tabexp is resetted */
-	tails = xnew(char *, sizeof(commands) / sizeof(struct command));
+	tails = xnew(char *, N_ELEMENTS(commands) - 1);
 	len = strlen(str);
 	pos = 0;
 	for (i = 0; commands[i].name; i++) {
@@ -2414,9 +2516,9 @@ static void expand_commands(const char *str)
 			free(tails[0]);
 			tails[0] = tmp;
 		}
-		tails[pos] = NULL;
 		tabexp.head = xstrdup(str);
 		tabexp.tails = tails;
+		tabexp.count = pos;
 	} else {
 		free(tails);
 	}
@@ -2503,7 +2605,7 @@ static void expand_command_line(const char *str)
 	tabexp.head = xstrdup(expbuf);
 }
 
-static void tab_expand(void)
+static void tab_expand(int direction)
 {
 	char *s1, *s2, *tmp;
 	int pos;
@@ -2519,7 +2621,7 @@ static void tab_expand(void)
 	/* tail */
 	s2 = xstrdup(cmdline.line + cmdline.bpos);
 
-	tmp = tabexp_expand(s1, expand_command_line);
+	tmp = tabexp_expand(s1, expand_command_line, direction);
 	if (tmp) {
 		/* tmp.s2 */
 		int l1, l2;
@@ -2534,8 +2636,8 @@ static void tab_expand(void)
 		}
 		sprintf(cmdline.line, "%s%s", tmp, s2);
 		cmdline.bpos = l1;
-		cmdline.cpos = u_strlen(tmp);
-		cmdline.clen = u_strlen(cmdline.line);
+		cmdline.cpos = u_strlen_safe(tmp);
+		cmdline.clen = u_strlen_safe(cmdline.line);
 		free(tmp);
 	}
 	free(s1);
@@ -2548,6 +2650,26 @@ static void reset_tab_expansion(void)
 	arg_expand_cmd = -1;
 }
 
+static void cmdline_modified(void)
+{
+	char *cmd, *arg;
+	struct command *c;
+
+	if (!parse_command(cmdline.line, &cmd, &arg))
+		return;
+
+	c = get_command(cmd);
+	if (!c)
+		goto end;
+
+	if (c->flags & CMD_LIVE)
+		run_parsed_command(cmd, arg);
+
+end:
+	free(cmd);
+	free(arg);
+}
+
 int parse_command(const char *buf, char **cmdp, char **argp)
 {
 	int cmd_start, cmd_end, cmd_len;
@@ -2665,6 +2787,7 @@ void command_mode_ch(uchar ch)
 		break;
 	case 0x04: // ^D
 		cmdline_delete_ch();
+		cmdline_modified();
 		break;
 	case 0x05: // ^E
 		cmdline_move_end();
@@ -2691,34 +2814,67 @@ void command_mode_ch(uchar ch)
 		break;
 	case 0x0B:
 		cmdline_clear_end();
+		cmdline_modified();
 		break;
 	case 0x09:
-		tab_expand();
+		tab_expand(1);
 		break;
 	case 0x15:
 		cmdline_backspace_to_bol();
+		cmdline_modified();
+		break;
+	case 0x17: // ^W
+		cmdline_backward_delete_word(cmdline_word_delimiters);
+		cmdline_modified();
 		break;
 	case 0x08: // ^H
 	case 127:
 		backspace();
+		cmdline_modified();
 		break;
 	default:
 		cmdline_insert_ch(ch);
+		cmdline_modified();
 	}
 	reset_history_search();
 	if (ch != 0x09)
 		reset_tab_expansion();
 }
 
+void command_mode_escape(int c)
+{
+	switch (c) {
+	case 98:
+		cmdline_backward_word(cmdline_filename_delimiters);
+		break;
+	case 100:
+		cmdline_delete_word(cmdline_filename_delimiters);
+		cmdline_modified();
+		break;
+	case 102:
+		cmdline_forward_word(cmdline_filename_delimiters);
+		break;
+	case 127:
+	case KEY_BACKSPACE:
+		cmdline_backward_delete_word(cmdline_filename_delimiters);
+		cmdline_modified();
+		break;
+	}
+	reset_history_search();
+}
+
 void command_mode_key(int key)
 {
-	reset_tab_expansion();
+	if (key != KEY_BTAB)
+		reset_tab_expansion();
 	switch (key) {
 	case KEY_DC:
 		cmdline_delete_ch();
+		cmdline_modified();
 		break;
 	case KEY_BACKSPACE:
 		backspace();
+		cmdline_modified();
 		break;
 	case KEY_LEFT:
 		cmdline_move_left();
@@ -2755,6 +2911,9 @@ void command_mode_key(int key)
 			}
 		}
 		return;
+	case KEY_BTAB:
+		tab_expand(-1);
+		break;
 	default:
 		d_print("key = %c (%d)\n", key, key);
 	}
@@ -2769,7 +2928,12 @@ void commands_init(void)
 
 void commands_exit(void)
 {
+	view_clear(TREE_VIEW);
+	view_clear(SORTED_VIEW);
+	view_clear(PLAYLIST_VIEW);
+	view_clear(QUEUE_VIEW);
 	history_save(&cmd_history);
+	history_free(&cmd_history);
 	free(cmd_history_filename);
 	tabexp_reset();
 }
