@@ -33,8 +33,21 @@
 #include "xmalloc.h"
 #include "debug.h"
 #include "utils.h"
+#include "config/ffmpeg.h"
 
-#define NUM_FFMPEG_KEYS 8
+#if (LIBAVFORMAT_VERSION_INT < ((52<<16)+(31<<8)+0))
+# define NUM_FFMPEG_KEYS 8
+#endif
+
+#if (LIBAVCODEC_VERSION_INT < ((52<<16)+(94<<8)+1))
+#define AV_SAMPLE_FMT_U8   SAMPLE_FMT_U8
+#define AV_SAMPLE_FMT_S16  SAMPLE_FMT_S16
+#define AV_SAMPLE_FMT_S32  SAMPLE_FMT_S32
+#define AV_SAMPLE_FMT_FLT  SAMPLE_FMT_FLT
+#if (LIBAVCODEC_VERSION_INT > ((51<<16)+(64<<8)+0))
+#define AV_SAMPLE_FMT_DBL  SAMPLE_FMT_DBL
+#endif
+#endif
 
 struct ffmpeg_input {
 	AVPacket pkt;
@@ -52,6 +65,7 @@ struct ffmpeg_output {
 struct ffmpeg_private {
 	AVCodecContext *codec_context;
 	AVFormatContext *input_context;
+	AVCodec *codec;
 	int stream_index;
 
 	struct ffmpeg_input *input;
@@ -74,6 +88,7 @@ static struct ffmpeg_input *ffmpeg_input_create(void)
 static void ffmpeg_input_free(struct ffmpeg_input *input)
 {
 	av_free_packet(&input->pkt);
+	free(input);
 }
 
 static struct ffmpeg_output *ffmpeg_output_create(void)
@@ -183,7 +198,7 @@ static int ffmpeg_open(struct input_plugin_data *ip_data)
 		codec = avcodec_find_decoder(cc->codec_id);
 		if (!codec) {
 			d_print("codec not found: %d, %s\n", cc->codec_id, cc->codec_name);
-			err = -IP_ERROR_FILE_FORMAT;
+			err = -IP_ERROR_UNSUPPORTED_FILE_TYPE;
 			break;
 		}
 
@@ -192,7 +207,16 @@ static int ffmpeg_open(struct input_plugin_data *ip_data)
 
 		if (avcodec_open(cc, codec) < 0) {
 			d_print("could not open codec: %d, %s\n", cc->codec_id, cc->codec_name);
-			err = -IP_ERROR_FILE_FORMAT;
+			err = -IP_ERROR_UNSUPPORTED_FILE_TYPE;
+			break;
+		}
+
+#if (LIBAVCODEC_VERSION_INT > ((51<<16)+(64<<8)+0))
+		if (cc->sample_fmt == AV_SAMPLE_FMT_FLT || cc->sample_fmt == AV_SAMPLE_FMT_DBL) {
+#else
+		if (cc->sample_fmt == AV_SAMPLE_FMT_FLT) {
+#endif
+			err = -IP_ERROR_SAMPLE_FORMAT;
 			break;
 		}
 		/* We assume below that no more errors follow. */
@@ -207,6 +231,7 @@ static int ffmpeg_open(struct input_plugin_data *ip_data)
 	priv = xnew(struct ffmpeg_private, 1);
 	priv->codec_context = cc;
 	priv->input_context = ic;
+	priv->codec = codec;
 	priv->stream_index = stream_index;
 	priv->input = ffmpeg_input_create();
 	if (priv->input == NULL) {
@@ -218,7 +243,22 @@ static int ffmpeg_open(struct input_plugin_data *ip_data)
 	priv->output = ffmpeg_output_create();
 
 	ip_data->private = priv;
-	ip_data->sf = sf_rate(cc->sample_rate) | sf_channels(cc->channels) | sf_bits(16) | sf_signed(1);
+	ip_data->sf = sf_rate(cc->sample_rate) | sf_channels(cc->channels);
+	switch (cc->sample_fmt) {
+	case AV_SAMPLE_FMT_U8:
+		ip_data->sf |= sf_bits(8) | sf_signed(0);
+		break;
+	case AV_SAMPLE_FMT_S32:
+		ip_data->sf |= sf_bits(32) | sf_signed(1);
+		break;
+	/* AV_SAMPLE_FMT_S16 */
+	default:
+		ip_data->sf |= sf_bits(16) | sf_signed(1);
+		break;
+	}
+#ifdef WORDS_BIGENDIAN
+	ip_data->sf |= sf_bigendian(1);
+#endif
 	return 0;
 }
 
@@ -242,12 +282,12 @@ static int ffmpeg_close(struct input_plugin_data *ip_data)
 static int ffmpeg_fill_buffer(AVFormatContext *ic, AVCodecContext *cc, struct ffmpeg_input *input,
 			      struct ffmpeg_output *output)
 {
-	/* frame_size specifies the size of output->buffer for
-	 * avcodec_decode_audio2. */
-	int frame_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;
-	int len;
-
 	while (1) {
+		/* frame_size specifies the size of output->buffer for
+		 * avcodec_decode_audio2. */
+		int frame_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;
+		int len;
+
 		if (input->curr_pkt_size <= 0) {
 			av_free_packet(&input->pkt);
 			if (av_read_frame(ic, &input->pkt) < 0) {
@@ -270,12 +310,14 @@ static int ffmpeg_fill_buffer(AVFormatContext *ic, AVCodecContext *cc, struct ff
 		len = avcodec_decode_audio2(cc, (int16_t *) output->buffer, &frame_size,
 				input->curr_pkt_buf, input->curr_pkt_size);
 #else
-		AVPacket avpkt;
-		av_init_packet(&avpkt);
-		avpkt.data = input->curr_pkt_buf;
-		avpkt.size = input->curr_pkt_size;
-		len = avcodec_decode_audio3(cc, (int16_t *) output->buffer, &frame_size, &avpkt);
-		av_free_packet(&avpkt);
+		{
+			AVPacket avpkt;
+			av_init_packet(&avpkt);
+			avpkt.data = input->curr_pkt_buf;
+			avpkt.size = input->curr_pkt_size;
+			len = avcodec_decode_audio3(cc, (int16_t *) output->buffer, &frame_size, &avpkt);
+			av_free_packet(&avpkt);
+		}
 #endif
 		if (len < 0) {
 			/* this is often reached when seeking, not sure why */
@@ -329,8 +371,7 @@ static int ffmpeg_seek(struct input_plugin_data *ip_data, double offset)
 #if (LIBAVFORMAT_VERSION_INT < ((50<<16)+(3<<8)+0))
 	int64_t pts = (int64_t) offset;
 #else
-	/* time base is 1/framerate */
-	int64_t pts = (int64_t) offset * st->time_base.den;
+	int64_t pts = av_rescale_q(offset * AV_TIME_BASE, AV_TIME_BASE_Q, st->time_base);
 #endif
 
 	ret = av_seek_frame(priv->input_context, priv->stream_index, pts, 0);
@@ -343,6 +384,7 @@ static int ffmpeg_seek(struct input_plugin_data *ip_data, double offset)
 	}
 }
 
+#if (LIBAVFORMAT_VERSION_INT < ((52<<16)+(31<<8)+0))
 /* Return new i. */
 static int set_comment(struct keyval *comment, int i, const char *key, const char *val)
 {
@@ -353,12 +395,15 @@ static int set_comment(struct keyval *comment, int i, const char *key, const cha
 	comment[i].val = xstrdup(val);
 	return i + 1;
 }
+#endif
 
 static int ffmpeg_read_comments(struct input_plugin_data *ip_data, struct keyval **comments)
 {
-	char buff[16];
 	struct ffmpeg_private *priv = ip_data->private;
 	AVFormatContext *ic = priv->input_context;
+
+#if (LIBAVFORMAT_VERSION_INT < ((52<<16)+(31<<8)+0))
+	char buff[16];
 	int i = 0;
 
 	*comments = xnew0(struct keyval, NUM_FFMPEG_KEYS + 1);
@@ -377,6 +422,18 @@ static int ffmpeg_read_comments(struct input_plugin_data *ip_data, struct keyval
 		snprintf(buff, sizeof(buff), "%d", ic->track);
 		i = set_comment(*comments, i, "tracknumber", buff);
 	}
+#else
+	GROWING_KEYVALS(c);
+	AVMetadataTag *tag = NULL;
+ 
+	while ((tag = av_metadata_get(ic->metadata, "", tag, AV_METADATA_IGNORE_SUFFIX))) {
+		if (tag && tag->value[0])
+			comments_add_const(&c, tag->key, tag->value);
+	}
+
+	keyvals_terminate(&c);
+	*comments = c.keyvals;
+#endif
 
 	return 0;
 }
@@ -384,7 +441,20 @@ static int ffmpeg_read_comments(struct input_plugin_data *ip_data, struct keyval
 static int ffmpeg_duration(struct input_plugin_data *ip_data)
 {
 	struct ffmpeg_private *priv = ip_data->private;
-	return priv->input_context->duration / 1000000L;
+	return priv->input_context->duration / AV_TIME_BASE;
+}
+
+static long ffmpeg_bitrate(struct input_plugin_data *ip_data)
+{
+	struct ffmpeg_private *priv = ip_data->private;
+	long bitrate = priv->input_context->bit_rate;
+	return bitrate ? bitrate : -IP_ERROR_FUNCTION_NOT_SUPPORTED;
+}
+
+static char *ffmpeg_codec(struct input_plugin_data *ip_data)
+{
+	struct ffmpeg_private *priv = ip_data->private;
+	return xstrdup(priv->codec->name);
 }
 
 const struct input_plugin_ops ip_ops = {
@@ -393,8 +463,21 @@ const struct input_plugin_ops ip_ops = {
 	.read = ffmpeg_read,
 	.seek = ffmpeg_seek,
 	.read_comments = ffmpeg_read_comments,
-	.duration = ffmpeg_duration
+	.duration = ffmpeg_duration,
+	.bitrate = ffmpeg_bitrate,
+	.codec = ffmpeg_codec
 };
 
-const char *const ip_extensions[] = { "wma", NULL };
+const int ip_priority = 30;
+#ifdef USE_FALLBACK_IP
+const char *const ip_extensions[] = { "any", NULL };
+#else
+const char *const ip_extensions[] = {
+	"ac3", "aif", "aifc", "aiff", "ape", "au", "mka", "shn", "tta", "wma",
+	/* also supported by other plugins */
+	"aac", "fla", "flac", "m4a", "m4b", "mp+", "mp2", "mp3", "mp4", "mpc",
+	"mpp", "ogg", "wav", "wv",
+	NULL
+};
+#endif
 const char *const ip_mime_types[] = { NULL };
