@@ -5,7 +5,7 @@
 #include "id3.h"
 #include "comment.h"
 #include "xmalloc.h"
-#include "utf8_encode.h"
+#include "convert.h"
 #include "uchar.h"
 #include "options.h"
 #include "debug.h"
@@ -16,8 +16,18 @@
 #include <errno.h>
 #include <stdio.h>
 #include <string.h>
+#include <strings.h>
 #include <limits.h>
 
+enum {
+	ID3_ENCODING_ISO_8859_1 = 0x00,
+	ID3_ENCODING_UTF_16     = 0x01,
+	ID3_ENCODING_UTF_16_BE  = 0x02,
+	ID3_ENCODING_UTF_8      = 0x03,
+
+	ID3_ENCODING_MAX        = 0x03
+};
+
 /*
  * position:
  *
@@ -255,13 +265,25 @@ static char *utf16_to_utf8(const unsigned char *buf, int buf_size)
 {
 	char *out;
 	int i, idx;
+	int little_endian = 0;
+
+	if (buf_size < 2)
+		return NULL;
+
+	if (buf[0] == 0xff && buf[1] == 0xfe)
+		little_endian = 1;
 
 	out = xnew(char, (buf_size / 2) * 4 + 1);
 	i = idx = 0;
+
 	while (buf_size - i >= 2) {
 		uchar u;
 
-		u = buf[i] + (buf[i + 1] << 8);
+		if (little_endian)
+			u = buf[i] + (buf[i + 1] << 8);
+		else
+			u = buf[i + 1] + (buf[i] << 8);
+
 		if (u_is_unicode(u)) {
 			if (!utf16_is_special(u))
 				u_set_char(out, &idx, u);
@@ -269,37 +291,13 @@ static char *utf16_to_utf8(const unsigned char *buf, int buf_size)
 			free(out);
 			return NULL;
 		}
+
 		if (u == 0)
 			return out;
+
 		i += 2;
 	}
-	u_set_char(out, &idx, 0);
-	return out;
-}
 
-static char *utf16be_to_utf8(const unsigned char *buf, int buf_size)
-{
-	char *out;
-	int i, idx;
-
-	out = xnew(char, (buf_size / 2) * 4 + 1);
-	i = 0;
-	idx = 0;
-	while (buf_size - i >= 2) {
-		uchar u;
-
-		u = buf[i + 1] + (buf[i] << 8);
-		if (u_is_unicode(u)) {
-		       if (!utf16_is_special(u))
-				u_set_char(out, &idx, u);
-		} else {
-			free(out);
-			return NULL;
-		}
-		if (u == 0)
-			return out;
-		i += 2;
-	}
 	u_set_char(out, &idx, 0);
 	return out;
 }
@@ -582,6 +580,46 @@ static int frame_tab_index(const char *id)
 	return -1;
 }
 
+static int check_date_format(const char *buf)
+{
+	int i, ch;
+
+	/* year */
+	for (i = 0; i < 4; i++) {
+		ch = *buf++;
+		if (ch < '0' || ch > '9')
+			return 0;
+	}
+	ch = *buf++;
+	if (!ch)
+		return 4;
+	if (ch != '-')
+		return 0;
+
+	/* month */
+	for (i = 0; i < 2; i++) {
+		ch = *buf++;
+		if (ch < '0' || ch > '9')
+			return 0;
+	}
+	ch = *buf++;
+	if (!ch)
+		return 7;
+	if (ch != '-')
+		return 0;
+
+	/* day */
+	for (i = 0; i < 2; i++) {
+		ch = *buf++;
+		if (ch < '0' || ch > '9')
+			return 0;
+	}
+	ch = *buf;
+	if (!ch || (ch >= '0' && ch <= '9'))
+		return 10;
+	return 0;
+}
+
 static void fix_date(char *buf)
 {
 	const char *ptr = buf;
@@ -609,12 +647,12 @@ static char *decode_str(const char *buf, int len, int encoding)
 	char *in, *out = NULL;
 
 	switch (encoding) {
-	case 0x00: /* ISO-8859-1 */
+	case ID3_ENCODING_ISO_8859_1:
 		in = xstrndup(buf, len);
 		utf8_encode(in, id3_default_charset, &out);
 		free(in);
 		break;
-	case 0x03: /* UTF-8 */
+	case ID3_ENCODING_UTF_8:
 		in = xstrndup(buf, len);
 		if (u_is_valid(in)) {
 			out = in;
@@ -623,12 +661,10 @@ static char *decode_str(const char *buf, int len, int encoding)
 			free(in);
 		}
 		break;
-	case 0x01: /* UTF-16 */
+	case ID3_ENCODING_UTF_16:
+	case ID3_ENCODING_UTF_16_BE:
 		out = utf16_to_utf8((const unsigned char *)buf, len);
 		break;
-	case 0x02: /* UTF-16BE */
-		out = utf16be_to_utf8((const unsigned char *)buf, len);
-		break;
 	}
 	return out;
 }
@@ -655,8 +691,12 @@ static void decode_normal(struct id3tag *id3, const char *buf, int len, int enco
 		free(out);
 		out = tmp;
 	} else if (key == ID3_DATE) {
+		int date_len = check_date_format(out);
 		id3_debug("date before: '%s'\n", out);
-		fix_date(out);
+		if (date_len)
+			out[date_len] = '\0';
+		else
+			fix_date(out);
 		if (!*out) {
 			id3_debug("date parsing failed\n");
 			free(out);
@@ -674,25 +714,22 @@ static void decode_normal(struct id3tag *id3, const char *buf, int len, int enco
 	add_v2(id3, key, out);
 }
 
-static int enc_is_utf16(int encoding)
+static size_t id3_skiplen(const char *buf, size_t len, int encoding)
 {
-	return encoding == 0x01 || encoding == 0x02;
-}
+	if (encoding == ID3_ENCODING_ISO_8859_1 || encoding == ID3_ENCODING_UTF_8) {
+		return strlen(buf) + 1;
+	} else {
+		int i = 0;
+		while (i + 1 < len) {
+			if (buf[i] == '\0' && buf[i + 1] == '\0')
+				return i + 2;
 
-static size_t id3_skiplen(const char *s, int encoding)
-{
-	size_t ret = strlen(s) + 1;
+			/* Assume every character is exactly 2 bytes */
+			i += 2;
+		}
 
-	if (enc_is_utf16(encoding)) {
-		/*
-		 * Assume BOM is always present,
-		 * and every character is 2 bytes long
-		 */
-		ret *= 2;
-		ret += 2;
+		BUG_ON(1);
 	}
-
-	return ret;
 }
 
 static void decode_txxx(struct id3tag *id3, const char *buf, int len, int encoding)
@@ -731,7 +768,7 @@ static void decode_txxx(struct id3tag *id3, const char *buf, int len, int encodi
 	else if (!strcasecmp(out, "compilation"))
 		key = ID3_COMPILATION;
 
-	size = id3_skiplen(out_mem, encoding);
+	size = id3_skiplen(buf, len, encoding);
 	free(out_mem);
 
 	if (key == NUM_ID3_KEYS)
@@ -751,14 +788,35 @@ static void decode_txxx(struct id3tag *id3, const char *buf, int len, int encodi
 
 static void decode_comment(struct id3tag *id3, const char *buf, int len, int encoding)
 {
+	int slen;
+	char out0;
 	char *out;
 
-	if (len <= 4)
+	if (len <= 3)
 		return;
 
 	/* skip language */
-	buf += 4;
-	len -= 4;
+	buf += 3;
+	len -= 3;
+
+	/* "Short content description" part of COMM frame */
+	out = decode_str(buf, len, encoding);
+	if (!out)
+		return;
+
+	out0 = *out;
+	free(out);
+
+	/* we are interested only in comments with empty description */
+	if (out0 != '\0')
+		return;
+
+	slen = id3_skiplen(buf, len, encoding);
+	if (slen >= len)
+		return;
+
+	buf += slen;
+	len -= slen;
 
 	out = decode_str(buf, len, encoding);
 	if (!out)
@@ -839,7 +897,7 @@ static void decode_rva2(struct id3tag *id3, const char *buf, int len)
 		return;
 	}
 
-	get_i16(buf, &volume_adj);
+	get_i16((unsigned char *)buf, &volume_adj);
 	buf += 2;
 
 	peak_bits = *buf++;
@@ -893,7 +951,7 @@ static void v2_add_frame(struct id3tag *id3, struct v2_frame_header *fh, const c
 	encoding = *buf++;
 	len = fh->size - 1;
 
-	if (encoding > 3)
+	if (encoding > ID3_ENCODING_MAX)
 		return;
 
 	idx = frame_tab_index(fh->id);
@@ -1003,7 +1061,7 @@ static int v2_read(struct id3tag *id3, int fd, const struct v2_header *header)
 		len_unsync = fh.size;
 
 		if ((fh.flags & V2_FRAME_UNSYNC) || (header->flags & V2_HEADER_UNSYNC))
-			unsync((unsigned char *)(buf + i), &fh.size);
+			unsync((unsigned char *)(buf + i), (int *)&fh.size);
 
 		v2_add_frame(id3, &fh, buf + i);
 
