@@ -36,6 +36,8 @@
 #include <stdarg.h>
 #include <math.h>
 
+const char * const player_status_names[] = { "stopped", "playing", "paused", NULL };
+
 enum producer_status {
 	PS_UNLOADED,
 	PS_STOPPED,
@@ -85,12 +87,14 @@ static sample_format_t buffer_sf;
 
 static pthread_t producer_thread;
 static pthread_mutex_t producer_mutex = CMUS_MUTEX_INITIALIZER;
+static pthread_cond_t producer_playing = CMUS_COND_INITIALIZER;
 static int producer_running = 1;
 static enum producer_status producer_status = PS_UNLOADED;
 static struct input_plugin *ip = NULL;
 
 static pthread_t consumer_thread;
 static pthread_mutex_t consumer_mutex = CMUS_MUTEX_INITIALIZER;
+static pthread_cond_t consumer_playing = CMUS_COND_INITIALIZER;
 static int consumer_running = 1;
 static enum consumer_status consumer_status = CS_STOPPED;
 static unsigned int consumer_pos = 0;
@@ -128,6 +132,7 @@ static void reset_buffer(void)
 	buffer_reset();
 	consumer_pos = 0;
 	scale_pos = 0;
+	pthread_cond_broadcast(&producer_playing);
 }
 
 static void set_buffer_sf(sample_format_t sf)
@@ -138,6 +143,9 @@ static void set_buffer_sf(sample_format_t sf)
 	if (sf_get_channels(buffer_sf) <= 2 && sf_get_bits(buffer_sf) <= 16) {
 		buffer_sf &= SF_RATE_MASK;
 		buffer_sf |= sf_channels(2) | sf_bits(16) | sf_signed(1);
+#ifdef WORDS_BIGENDIAN
+		buffer_sf |= sf_bigendian(1);
+#endif
 	}
 }
 
@@ -162,43 +170,124 @@ static const unsigned short soft_vol_db[100] = {
 	0xcdf1, 0xd71a, 0xe59c, 0xefd3
 };
 
-#ifdef swap16   /* e.g. OpenBSD */
-#undef swap16
-#endif
-static inline unsigned short swap16(unsigned short u)
+static inline void scale_sample_int16_t(int16_t *buf, int i, int vol, int swap)
 {
-	return (u << 8) | (u >> 8);
-}
-
-static inline void scale_sample(signed short *buf, int i, int vol)
-{
-#ifdef WORDS_BIGENDIAN
-	int sample = (short)swap16(buf[i]);
-#else
-	int sample = buf[i];
-#endif
+	int32_t sample = swap ? (int16_t)swap_uint16(buf[i]) : buf[i];
 
 	if (sample < 0) {
 		sample = (sample * vol - SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
-		if (sample < -32768)
-			sample = -32768;
+		if (sample < INT16_MIN)
+			sample = INT16_MIN;
 	} else {
 		sample = (sample * vol + SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
-		if (sample > 32767)
-			sample = 32767;
+		if (sample > INT16_MAX)
+			sample = INT16_MAX;
 	}
+	buf[i] = swap ? swap_uint16(sample) : sample;
+}
+
+static inline int32_t scale_sample_s24le(int32_t s, int vol)
+{
+	int64_t sample = s;
+	if (sample < 0) {
+		sample = (sample * vol - SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample < -0x800000)
+			sample = -0x800000;
+	} else {
+		sample = (sample * vol + SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample > 0x7fffff)
+			sample = 0x7fffff;
+	}
+	return sample;
+}
+
+static inline void scale_sample_int32_t(int32_t *buf, int i, int vol, int swap)
+{
+	int64_t sample = swap ? (int32_t)swap_uint32(buf[i]) : buf[i];
+
+	if (sample < 0) {
+		sample = (sample * vol - SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample < INT32_MIN)
+			sample = INT32_MIN;
+	} else {
+		sample = (sample * vol + SOFT_VOL_SCALE / 2) / SOFT_VOL_SCALE;
+		if (sample > INT32_MAX)
+			sample = INT32_MAX;
+	}
+	buf[i] = swap ? swap_uint32(sample) : sample;
+}
+
+static inline int sf_need_swap(sample_format_t sf)
+{
 #ifdef WORDS_BIGENDIAN
-	buf[i] = swap16(sample);
+	return !sf_get_bigendian(sf);
 #else
-	buf[i] = sample;
+	return sf_get_bigendian(sf);
 #endif
 }
 
+#define SCALE_SAMPLES(TYPE, buffer, count, l, r, swap)				\
+{										\
+	const int frames = count / sizeof(TYPE) / 2;				\
+	TYPE *buf = (TYPE *) buffer;						\
+	int i;									\
+	/* avoid underflowing -32768 to 32767 when scale is 65536 */		\
+	if (l != SOFT_VOL_SCALE && r != SOFT_VOL_SCALE) {			\
+		for (i = 0; i < frames; i++) {					\
+			scale_sample_##TYPE(buf, i * 2, l, swap);		\
+			scale_sample_##TYPE(buf, i * 2 + 1, r, swap);		\
+		}								\
+	} else if (l != SOFT_VOL_SCALE) {					\
+		for (i = 0; i < frames; i++)					\
+			scale_sample_##TYPE(buf, i * 2, l, swap);		\
+	} else if (r != SOFT_VOL_SCALE) {					\
+		for (i = 0; i < frames; i++)					\
+			scale_sample_##TYPE(buf, i * 2 + 1, r, swap);		\
+	}									\
+}
+
+static inline int32_t read_s24le(const char *buf)
+{
+	const unsigned char *b = (const unsigned char *) buf;
+	return b[0] | (b[1] << 8) | (((const signed char *) buf)[2] << 16);
+}
+
+static inline void write_s24le(char *buf, int32_t x)
+{
+	unsigned char *b = (unsigned char *) buf;
+	b[0] = x;
+	b[1] = x >> 8;
+	b[2] = x >> 16;
+}
+
+static void scale_samples_s24le(char *buf, unsigned int count, int l, int r)
+{
+	int frames = count / 3 / 2;
+	if (l != SOFT_VOL_SCALE && r != SOFT_VOL_SCALE) {
+		while (frames--) {
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), l));
+			buf += 3;
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), r));
+			buf += 3;
+		}
+	} else if (l != SOFT_VOL_SCALE) {
+		while (frames--) {
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), l));
+			buf += 3 * 2;
+		}
+	} else if (r != SOFT_VOL_SCALE) {
+		buf += 3;
+		while (frames--) {
+			write_s24le(buf, scale_sample_s24le(read_s24le(buf), r));
+			buf += 3 * 2;
+		}
+	}
+}
+
 static void scale_samples(char *buffer, unsigned int *countp)
 {
-	signed short *buf;
 	unsigned int count = *countp;
-	int ch, bits, l, r, i;
+	int ch, bits, l, r;
 
 	BUG_ON(scale_pos < consumer_pos);
 
@@ -211,14 +300,13 @@ static void scale_samples(char *buffer, unsigned int *countp)
 		count -= offs;
 	}
 	scale_pos += count;
-	buf = (signed short *)buffer;
 
 	if (replaygain_scale == 1.0 && soft_vol_l == 100 && soft_vol_r == 100)
 		return;
 
 	ch = sf_get_channels(buffer_sf);
 	bits = sf_get_bits(buffer_sf);
-	if (ch != 2 || bits != 16)
+	if (ch != 2 || (bits != 16 && bits != 24 && bits != 32))
 		return;
 
 	l = SOFT_VOL_SCALE;
@@ -231,32 +319,22 @@ static void scale_samples(char *buffer, unsigned int *countp)
 	l *= replaygain_scale;
 	r *= replaygain_scale;
 
-	/* avoid underflowing -32768 to 32767 when scale is 65536 */
-	if (l != SOFT_VOL_SCALE && r != SOFT_VOL_SCALE) {
-		for (i = 0; i < count / 4; i++) {
-			scale_sample(buf, i * 2, l);
-			scale_sample(buf, i * 2 + 1, r);
-		}
-	} else if (l != SOFT_VOL_SCALE) {
-		for (i = 0; i < count / 4; i++)
-			scale_sample(buf, i * 2, l);
-	} else if (r != SOFT_VOL_SCALE) {
-		for (i = 0; i < count / 4; i++)
-			scale_sample(buf, i * 2 + 1, r);
+	switch (bits) {
+	case 16:
+		SCALE_SAMPLES(int16_t, buffer, count, l, r, sf_need_swap(buffer_sf));
+		break;
+	case 24:
+		if (likely(!sf_get_bigendian(buffer_sf)))
+			scale_samples_s24le(buffer, count, l, r);
+		break;
+	case 32:
+		SCALE_SAMPLES(int32_t, buffer, count, l, r, sf_need_swap(buffer_sf));
+		break;
 	}
 }
 
-static int parse_double(const char *str, double *val)
-{
-	char *end;
-
-	*val = strtod(str, &end);
-	return str == end;
-}
-
 static void update_rg_scale(void)
 {
-	const char *g, *p;
 	double gain, peak, db, scale, limit;
 
 	replaygain_scale = 1.0;
@@ -264,32 +342,28 @@ static void update_rg_scale(void)
 		return;
 
 	if (replaygain == RG_TRACK || replaygain == RG_TRACK_PREFERRED) {
-		g = keyvals_get_val(player_info.ti->comments, "replaygain_track_gain");
-		p = keyvals_get_val(player_info.ti->comments, "replaygain_track_peak");
+		gain = player_info.ti->rg_track_gain;
+		peak = player_info.ti->rg_track_peak;
 	} else {
-		g = keyvals_get_val(player_info.ti->comments, "replaygain_album_gain");
-		p = keyvals_get_val(player_info.ti->comments, "replaygain_album_peak");
+		gain = player_info.ti->rg_album_gain;
+		peak = player_info.ti->rg_album_peak;
 	}
 
-	if (!g || !p) {
+	if (isnan(gain)) {
 		if (replaygain == RG_TRACK_PREFERRED) {
-			g = keyvals_get_val(player_info.ti->comments, "replaygain_album_gain");
-			p = keyvals_get_val(player_info.ti->comments, "replaygain_album_peak");
+			gain = player_info.ti->rg_album_gain;
+			peak = player_info.ti->rg_album_peak;
 		} else if (replaygain == RG_ALBUM_PREFERRED) {
-			g = keyvals_get_val(player_info.ti->comments, "replaygain_track_gain");
-			p = keyvals_get_val(player_info.ti->comments, "replaygain_track_peak");
+			gain = player_info.ti->rg_track_gain;
+			peak = player_info.ti->rg_track_peak;
 		}
 	}
 
-	if (!g || !p) {
-		d_print("gain or peak not available\n");
+	if (isnan(gain)) {
+		d_print("gain not available\n");
 		return;
 	}
-	if (parse_double(g, &gain) || parse_double(p, &peak)) {
-		d_print("could not parse gain (%s) or peak (%s)\n", g, p);
-		return;
-	}
-	if (peak < 0.05) {
+	if (!isnan(peak) && peak < 0.05) {
 		d_print("peak (%g) is too small\n", peak);
 		return;
 	}
@@ -299,8 +373,10 @@ static void update_rg_scale(void)
 	scale = pow(10.0, db / 20.0);
 	replaygain_scale = scale;
 	limit = 1.0 / peak;
-	if (replaygain_limit && replaygain_scale > limit)
-		replaygain_scale = limit;
+	if (replaygain_limit && !isnan(peak)) {
+		if (replaygain_scale > limit)
+			replaygain_scale = limit;
+	}
 
 	d_print("gain = %f, peak = %f, db = %f, scale = %f, limit = %f, replaygain_scale = %f\n",
 			gain, peak, db, scale, limit, replaygain_scale);
@@ -352,7 +428,7 @@ static inline void metadata_changed(void)
 	if (!rc) {
 		if (player_info.ti->comments)
 			keyvals_free(player_info.ti->comments);
-		player_info.ti->comments = comments;
+		track_info_set_comments(player_info.ti, comments);
 	}
 
 	player_info.metadata_changed = 1;
@@ -523,6 +599,12 @@ static void __prebuffer(void)
 
 /* setting producer status {{{ */
 
+static void __producer_status_update(enum producer_status status)
+{
+	producer_status =  status;
+	pthread_cond_broadcast(&producer_playing);
+}
+
 static void __producer_play(void)
 {
 	if (producer_status == PS_UNLOADED) {
@@ -540,7 +622,7 @@ static void __producer_play(void)
 				file_changed(NULL);
 			} else {
 				ip_setup(ip);
-				producer_status = PS_PLAYING;
+				__producer_status_update(PS_PLAYING);
 				file_changed(ti);
 			}
 		}
@@ -555,13 +637,13 @@ static void __producer_play(void)
 		if (rc) {
 			player_ip_error(rc, "opening file `%s'", ip_get_filename(ip));
 			ip_delete(ip);
-			producer_status = PS_UNLOADED;
+			__producer_status_update(PS_UNLOADED);
 		} else {
 			ip_setup(ip);
-			producer_status = PS_PLAYING;
+			__producer_status_update(PS_PLAYING);
 		}
 	} else if (producer_status == PS_PAUSED) {
-		producer_status = PS_PLAYING;
+		__producer_status_update(PS_PLAYING);
 	}
 }
 
@@ -569,7 +651,7 @@ static void __producer_stop(void)
 {
 	if (producer_status == PS_PLAYING || producer_status == PS_PAUSED) {
 		ip_close(ip);
-		producer_status = PS_STOPPED;
+		__producer_status_update(PS_STOPPED);
 		reset_buffer();
 	}
 }
@@ -579,16 +661,16 @@ static void __producer_unload(void)
 	__producer_stop();
 	if (producer_status == PS_STOPPED) {
 		ip_delete(ip);
-		producer_status = PS_UNLOADED;
+		__producer_status_update(PS_UNLOADED);
 	}
 }
 
 static void __producer_pause(void)
 {
 	if (producer_status == PS_PLAYING) {
-		producer_status = PS_PAUSED;
+		__producer_status_update(PS_PAUSED);
 	} else if (producer_status == PS_PAUSED) {
-		producer_status = PS_PLAYING;
+		__producer_status_update(PS_PLAYING);
 	}
 }
 
@@ -596,7 +678,7 @@ static void __producer_set_file(struct track_info *ti)
 {
 	__producer_unload();
 	ip = ip_new(ti->filename);
-	producer_status = PS_STOPPED;
+	__producer_status_update(PS_STOPPED);
 	file_changed(ti);
 }
 
@@ -604,6 +686,12 @@ static void __producer_set_file(struct track_info *ti)
 
 /* setting consumer status {{{ */
 
+static void __consumer_status_update(enum consumer_status status)
+{
+	consumer_status = status;
+	pthread_cond_broadcast(&consumer_playing);
+}
+
 static void __consumer_play(void)
 {
 	if (consumer_status == CS_PLAYING) {
@@ -616,11 +704,11 @@ static void __consumer_play(void)
 		if (rc) {
 			player_op_error(rc, "opening audio device");
 		} else {
-			consumer_status = CS_PLAYING;
+			__consumer_status_update(CS_PLAYING);
 		}
 	} else if (consumer_status == CS_PAUSED) {
 		op_unpause();
-		consumer_status = CS_PLAYING;
+		__consumer_status_update(CS_PLAYING);
 	}
 }
 
@@ -628,7 +716,7 @@ static void __consumer_drain_and_stop(void)
 {
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
 		op_close();
-		consumer_status = CS_STOPPED;
+		__consumer_status_update(CS_STOPPED);
 	}
 }
 
@@ -637,7 +725,7 @@ static void __consumer_stop(void)
 	if (consumer_status == CS_PLAYING || consumer_status == CS_PAUSED) {
 		op_drop();
 		op_close();
-		consumer_status = CS_STOPPED;
+		__consumer_status_update(CS_STOPPED);
 	}
 }
 
@@ -645,10 +733,10 @@ static void __consumer_pause(void)
 {
 	if (consumer_status == CS_PLAYING) {
 		op_pause();
-		consumer_status = CS_PAUSED;
+		__consumer_status_update(CS_PAUSED);
 	} else if (consumer_status == CS_PAUSED) {
 		op_unpause();
-		consumer_status = CS_PLAYING;
+		__consumer_status_update(CS_PLAYING);
 	}
 }
 
@@ -669,7 +757,7 @@ static int change_sf(sample_format_t sf, int drop)
 		rc = op_open(buffer_sf);
 		if (rc) {
 			player_op_error(rc, "opening audio device");
-			consumer_status = CS_STOPPED;
+			__consumer_status_update(CS_STOPPED);
 			__producer_stop();
 			return rc;
 		}
@@ -677,7 +765,7 @@ static int change_sf(sample_format_t sf, int drop)
 		op_drop();
 		op_unpause();
 	}
-	consumer_status = CS_PLAYING;
+	__consumer_status_update(CS_PLAYING);
 	return 0;
 }
 
@@ -707,7 +795,7 @@ static void __consumer_handle_eof(void)
 	if (get_next(&ti) == 0) {
 		__producer_unload();
 		ip = ip_new(ti->filename);
-		producer_status = PS_STOPPED;
+		__producer_status_update(PS_STOPPED);
 		/* PS_STOPPED, CS_PLAYING */
 		if (player_cont) {
 			__producer_play();
@@ -745,8 +833,8 @@ static void *consumer_loop(void *arg)
 			break;
 
 		if (consumer_status == CS_PAUSED || consumer_status == CS_STOPPED) {
+			pthread_cond_wait(&consumer_playing, &consumer_mutex);
 			consumer_unlock();
-			ms_sleep(50);
 			continue;
 		}
 		space = op_buffer_space();
@@ -756,7 +844,7 @@ static void *consumer_loop(void *arg)
 
 			/* try to reopen */
 			op_close();
-			consumer_status = CS_STOPPED;
+			__consumer_status_update(CS_STOPPED);
 			__consumer_play();
 
 			consumer_unlock();
@@ -806,7 +894,7 @@ static void *consumer_loop(void *arg)
 			if (size > space)
 				size = space;
 			if (soft_vol || replaygain)
-				scale_samples(rpos, &size);
+				scale_samples(rpos, (unsigned int *)&size);
 			rc = op_write(rpos, size);
 			if (rc < 0) {
 				d_print("op_write returned %d %s\n", rc,
@@ -814,7 +902,7 @@ static void *consumer_loop(void *arg)
 
 				/* try to reopen */
 				op_close();
-				consumer_status = CS_STOPPED;
+				__consumer_status_update(CS_STOPPED);
 				__consumer_play();
 
 				consumer_unlock();
@@ -848,8 +936,8 @@ static void *producer_loop(void *arg)
 		if (producer_status == PS_UNLOADED ||
 		    producer_status == PS_PAUSED ||
 		    producer_status == PS_STOPPED || ip_eof(ip)) {
+			pthread_cond_wait(&producer_playing, &producer_mutex);
 			producer_unlock();
-			ms_sleep(50);
 			continue;
 		}
 		for (i = 0; ; i++) {
@@ -953,13 +1041,16 @@ void player_exit(void)
 
 	player_lock();
 	consumer_running = 0;
+	pthread_cond_broadcast(&consumer_playing);
 	producer_running = 0;
+	pthread_cond_broadcast(&producer_playing);
 	player_unlock();
 	
 	rc = pthread_join(consumer_thread, NULL);
 	BUG_ON(rc);
 	rc = pthread_join(producer_thread, NULL);
 	BUG_ON(rc);
+	buffer_free();
 }
 
 void player_stop(void)
@@ -1085,10 +1176,12 @@ out:
 	player_unlock();
 }
 
-void player_seek(double offset, int relative)
+void player_seek(double offset, int relative, int start_playing)
 {
+	int stopped = 0;
 	player_lock();
 	if (consumer_status == CS_STOPPED) {
+		stopped = 1;
 		__producer_play();
 		if (producer_status == PS_PLAYING) {
 			__consumer_play();
@@ -1145,12 +1238,17 @@ void player_seek(double offset, int relative)
 /* 		d_print("seeking %g/%g (%g from eof)\n", new_pos, duration, duration - new_pos); */
 		rc = ip_seek(ip, new_pos);
 		if (rc == 0) {
-/* 			d_print("doing op_drop after seek\n"); */
+			d_print("doing op_drop after seek\n");
 			op_drop();
 			reset_buffer();
 			consumer_pos = new_pos * buffer_second_size();
 			scale_pos = consumer_pos;
 			__consumer_position_update();
+			if (stopped && !start_playing) {
+				__producer_pause();
+				__consumer_pause();
+				__player_status_changed();
+			}
 		} else {
 			d_print("error: ip_seek returned %d\n", rc);
 		}
@@ -1183,7 +1281,7 @@ void player_set_op(const char *name)
 		rc = op_select_any();
 	}
 	if (rc) {
-		consumer_status = CS_STOPPED;
+		__consumer_status_update(CS_STOPPED);
 
 		__producer_stop();
 		if (name)
@@ -1198,7 +1296,7 @@ void player_set_op(const char *name)
 		set_buffer_sf(ip_get_sf(ip));
 		rc = op_open(buffer_sf);
 		if (rc) {
-			consumer_status = CS_STOPPED;
+			__consumer_status_update(CS_STOPPED);
 			__producer_stop();
 			player_op_error(rc, "opening audio device");
 			player_unlock();
