@@ -45,7 +45,9 @@
 #include "help.h"
 #include "worker.h"
 #include "input.h"
+#include "file.h"
 #include "config/curses.h"
+#include "config/iconv.h"
 
 #include <unistd.h>
 #include <stdlib.h>
@@ -56,9 +58,12 @@
 #include <dirent.h>
 #include <locale.h>
 #include <langinfo.h>
+#ifdef HAVE_ICONV
 #include <iconv.h>
+#endif
 #include <signal.h>
 #include <stdarg.h>
+#include <math.h>
 
 #if defined(__sun__) || defined(__CYGWIN__)
 /* TIOCGWINSZ */
@@ -80,12 +85,20 @@ enum ui_input_mode input_mode = NORMAL_MODE;
 int cur_view = TREE_VIEW;
 struct searchable *searchable;
 char *lib_filename = NULL;
+char *lib_ext_filename = NULL;
 char *pl_filename = NULL;
+char *pl_ext_filename = NULL;
+char *play_queue_filename = NULL;
+char *play_queue_ext_filename = NULL;
+char *charset = NULL;
+int using_utf8 = 0;
+
 
 /* ------------------------------------------------------------------------- */
 
 static char *lib_autosave_filename;
 static char *pl_autosave_filename;
+static char *play_queue_autosave_filename;
 
 /* shown error message and time stamp
  * error is cleared if it is older than 3s and key was pressed
@@ -98,7 +111,6 @@ static int error_count = 0;
 
 static char *server_address = NULL;
 
-static char *charset = NULL;
 static char print_buffer[512];
 
 /* destination buffer for utf8_encode and utf8_decode */
@@ -107,8 +119,10 @@ static char conv_buffer[512];
 /* one character can take up to 4 bytes in UTF-8 */
 #define print_buffer_max_width (sizeof(print_buffer) / 4 - 1)
 
-static int using_utf8;
+/* used for messages to the client */
+static int client_fd = -1;
 
+static char tcap_buffer[64];
 static const char *t_ts;
 static const char *t_fs;
 
@@ -206,24 +220,36 @@ enum {
 	TF_GENRE,
 	TF_COMMENT,
 	TF_DURATION,
+	TF_BITRATE,
+	TF_CODEC,
 	TF_PATHFILE,
 	TF_FILE,
+	TF_RG_TRACK_GAIN,
+	TF_RG_TRACK_PEAK,
+	TF_RG_ALBUM_GAIN,
+	TF_RG_ALBUM_PEAK,
 	NR_TFS
 };
 
 static struct format_option track_fopts[NR_TFS + 1] = {
-	DEF_FO_STR('A'),
-	DEF_FO_STR('a'),
-	DEF_FO_STR('l'),
-	DEF_FO_INT('D'),
-	DEF_FO_INT('n'),
-	DEF_FO_STR('t'),
-	DEF_FO_STR('y'),
-	DEF_FO_STR('g'),
-	DEF_FO_STR('c'),
-	DEF_FO_TIME('d'),
-	DEF_FO_STR('f'),
-	DEF_FO_STR('F'),
+	DEF_FO_STR('A', "albumartist", 0),
+	DEF_FO_STR('a', "artist", 0),
+	DEF_FO_STR('l', "album", 0),
+	DEF_FO_INT('D', "discnumber", 1),
+	DEF_FO_INT('n', "tracknumber", 1),
+	DEF_FO_STR('t', "title", 0),
+	DEF_FO_STR('y', "date", 1),
+	DEF_FO_STR('g', "genre", 0),
+	DEF_FO_STR('c', "comment", 0),
+	DEF_FO_TIME('d', "duration", 0),
+	DEF_FO_INT('\0', "bitrate", 0),
+	DEF_FO_STR('\0', "codec", 0),
+	DEF_FO_STR('f', "path", 0),
+	DEF_FO_STR('F', "filename", 0),
+	DEF_FO_DOUBLE('\0', "rg_track_gain", 0),
+	DEF_FO_DOUBLE('\0', "rg_track_peak", 0),
+	DEF_FO_DOUBLE('\0', "rg_album_gain", 0),
+	DEF_FO_DOUBLE('\0', "rg_album_peak", 0),
 	DEF_FO_END
 };
 
@@ -244,23 +270,30 @@ enum {
 };
 
 static struct format_option status_fopts[NR_SFS + 1] = {
-	DEF_FO_STR('s'),
-	DEF_FO_TIME('p'),
-	DEF_FO_TIME('d'),
-	DEF_FO_TIME('t'),
-	DEF_FO_INT('v'),
-	DEF_FO_INT('l'),
-	DEF_FO_INT('r'),
-	DEF_FO_INT('b'),
-	DEF_FO_STR('R'),
-	DEF_FO_STR('C'),
-	DEF_FO_STR('S'),
-	DEF_FO_STR('L'),
+	DEF_FO_STR('s', NULL, 0),
+	DEF_FO_TIME('p', NULL, 0),
+	DEF_FO_TIME('d', NULL, 0),
+	DEF_FO_TIME('t', NULL, 0),
+	DEF_FO_INT('v', NULL, 0),
+	DEF_FO_INT('l', NULL, 0),
+	DEF_FO_INT('r', NULL, 0),
+	DEF_FO_INT('b', NULL, 0),
+	DEF_FO_STR('R', NULL, 0),
+	DEF_FO_STR('C', NULL, 0),
+	DEF_FO_STR('S', NULL, 0),
+	DEF_FO_STR('L', NULL, 0),
 	DEF_FO_END
 };
 
+int track_format_valid(const char *format)
+{
+	return format_valid(format, track_fopts);
+}
+
 static void utf8_encode(const char *buffer)
 {
+	int n;
+#ifdef HAVE_ICONV
 	static iconv_t cd = (iconv_t)-1;
 	size_t is, os;
 	const char *i;
@@ -272,7 +305,7 @@ static void utf8_encode(const char *buffer)
 		cd = iconv_open("UTF-8", charset);
 		if (cd == (iconv_t)-1) {
 			d_print("iconv_open failed: %s\n", strerror(errno));
-			return;
+			goto fallback;
 		}
 	}
 	i = buffer;
@@ -283,12 +316,20 @@ static void utf8_encode(const char *buffer)
 	*o = 0;
 	if (rc == -1) {
 		d_print("iconv failed: %s\n", strerror(errno));
-		return;
+		goto fallback;
 	}
+	return;
+fallback:
+#endif
+	n = min(sizeof(conv_buffer) - 1, strlen(buffer));
+	memmove(conv_buffer, buffer, n);
+	conv_buffer[n] = '\0';
 }
 
 static void utf8_decode(const char *buffer)
 {
+	int n;
+#ifdef HAVE_ICONV
 	static iconv_t cd = (iconv_t)-1;
 	size_t is, os;
 	const char *i;
@@ -300,7 +341,7 @@ static void utf8_decode(const char *buffer)
 		cd = iconv_open(charset, "UTF-8");
 		if (cd == (iconv_t)-1) {
 			d_print("iconv_open failed: %s\n", strerror(errno));
-			return;
+			goto fallback;
 		}
 	}
 	i = buffer;
@@ -311,8 +352,13 @@ static void utf8_decode(const char *buffer)
 	*o = 0;
 	if (rc == -1) {
 		d_print("iconv failed: %s\n", strerror(errno));
-		return;
+		goto fallback;
 	}
+	return;
+fallback:
+#endif
+	n = u_to_ascii(conv_buffer, buffer, sizeof(conv_buffer) - 1);
+	conv_buffer[n] = '\0';
 }
 
 /* screen updates {{{ */
@@ -320,10 +366,10 @@ static void utf8_decode(const char *buffer)
 static void dump_print_buffer(int row, int col)
 {
 	if (using_utf8) {
-		mvaddstr(row, col, print_buffer);
+		(void) mvaddstr(row, col, print_buffer);
 	} else {
 		utf8_decode(print_buffer);
-		mvaddstr(row, col, conv_buffer);
+		(void) mvaddstr(row, col, conv_buffer);
 	}
 }
 
@@ -340,7 +386,7 @@ static int format_str(char *buf, const char *str, int width)
 		uchar u;
 		int w;
 
-		u_get_char(str, &s, &u);
+		u = u_get_char(str, &s);
 		if (u == 0) {
 			memset(buf + d, ' ', width);
 			d += width;
@@ -404,7 +450,7 @@ static void sprint_ascii(int row, int col, const char *str, int len)
 	}
 	print_buffer[len + 1] = ' ';
 	print_buffer[len + 2] = 0;
-	mvaddstr(row, col, print_buffer);
+	(void) mvaddstr(row, col, print_buffer);
 }
 
 static void print_tree(struct window *win, int row, struct iter *iter)
@@ -471,6 +517,13 @@ static inline void fopt_set_int(struct format_option *fopt, int value, int empty
 	fopt->empty = empty;
 }
 
+static inline void fopt_set_double(struct format_option *fopt, double value, int empty)
+{
+	BUG_ON(fopt->type != FO_DOUBLE);
+	fopt->fo_double = value;
+	fopt->empty = empty;
+}
+
 static inline void fopt_set_time(struct format_option *fopt, int value, int empty)
 {
 	BUG_ON(fopt->type != FO_TIME);
@@ -481,7 +534,6 @@ static inline void fopt_set_time(struct format_option *fopt, int value, int empt
 static void fill_track_fopts_track_info(struct track_info *info)
 {
 	char *filename;
-	int num, disc;
 
 	if (using_utf8) {
 		filename = info->filename;
@@ -489,19 +541,23 @@ static void fill_track_fopts_track_info(struct track_info *info)
 		utf8_encode(info->filename);
 		filename = conv_buffer;
 	}
-	disc = comments_get_int(info->comments, "discnumber");
-	num = comments_get_int(info->comments, "tracknumber");
 
-	fopt_set_str(&track_fopts[TF_ALBUMARTIST], comments_get_albumartist(info->comments));
-	fopt_set_str(&track_fopts[TF_ARTIST], keyvals_get_val(info->comments, "artist"));
-	fopt_set_str(&track_fopts[TF_ALBUM], keyvals_get_val(info->comments, "album"));
-	fopt_set_int(&track_fopts[TF_DISC], disc, disc == -1);
-	fopt_set_int(&track_fopts[TF_TRACK], num, num == -1);
-	fopt_set_str(&track_fopts[TF_TITLE], keyvals_get_val(info->comments, "title"));
+	fopt_set_str(&track_fopts[TF_ALBUMARTIST], info->albumartist);
+	fopt_set_str(&track_fopts[TF_ARTIST], info->artist);
+	fopt_set_str(&track_fopts[TF_ALBUM], info->album);
+	fopt_set_int(&track_fopts[TF_DISC], info->discnumber, info->discnumber == -1);
+	fopt_set_int(&track_fopts[TF_TRACK], info->tracknumber, info->tracknumber == -1);
+	fopt_set_str(&track_fopts[TF_TITLE], info->title);
 	fopt_set_str(&track_fopts[TF_YEAR], keyvals_get_val(info->comments, "date"));
-	fopt_set_str(&track_fopts[TF_GENRE], keyvals_get_val(info->comments, "genre"));
-	fopt_set_str(&track_fopts[TF_COMMENT], keyvals_get_val(info->comments, "comment"));
+	fopt_set_str(&track_fopts[TF_GENRE], info->genre);
+	fopt_set_str(&track_fopts[TF_COMMENT], info->comment);
 	fopt_set_time(&track_fopts[TF_DURATION], info->duration, info->duration == -1);
+	fopt_set_double(&track_fopts[TF_RG_TRACK_GAIN], info->rg_track_gain, isnan(info->rg_track_gain));
+	fopt_set_double(&track_fopts[TF_RG_TRACK_PEAK], info->rg_track_peak, isnan(info->rg_track_peak));
+	fopt_set_double(&track_fopts[TF_RG_ALBUM_GAIN], info->rg_album_gain, isnan(info->rg_album_gain));
+	fopt_set_double(&track_fopts[TF_RG_ALBUM_PEAK], info->rg_album_peak, isnan(info->rg_album_peak));
+	fopt_set_int(&track_fopts[TF_BITRATE], (int) (info->bitrate / 1000. + 0.5), info->bitrate == -1);
+	fopt_set_str(&track_fopts[TF_CODEC], info->codec);
 	fopt_set_str(&track_fopts[TF_PATHFILE], filename);
 	if (is_url(info->filename)) {
 		fopt_set_str(&track_fopts[TF_FILE], filename);
@@ -644,6 +700,7 @@ static void print_filter(struct window *win, int row, struct iter *iter)
 	int current = !!e->act_stat;
 	const char stat_chars[3] = " *!";
 	int ch1, ch2, ch3, pos;
+	const char *e_filter;
 
 	window_get_sel(win, &sel);
 	selected = iters_equal(iter, &sel);
@@ -661,7 +718,14 @@ static void print_filter(struct window *win, int row, struct iter *iter)
 		ch3 = ']';
 	}
 	ch2 = stat_chars[e->sel_stat];
-	snprintf(buf, sizeof(buf), "%c%c%c%-15s  %s", ch1, ch2, ch3, e->name, e->filter);
+
+	e_filter = e->filter;
+	if (!using_utf8) {
+		utf8_encode(e_filter);
+		e_filter = conv_buffer;
+	}
+
+	snprintf(buf, sizeof(buf), "%c%c%c%-15s  %s", ch1, ch2, ch3, e->name, e_filter);
 	pos = format_str(print_buffer, buf, COLS - 1);
 	print_buffer[pos++] = ' ';
 	print_buffer[pos] = 0;
@@ -861,10 +925,10 @@ static void draw_separator(void)
 	int row;
 
 	bkgdset(pairs[CURSED_WIN_TITLE]);
-	mvaddch(0, tree_win_w, ' ');
+	(void) mvaddch(0, tree_win_w, ' ');
 	bkgdset(pairs[CURSED_SEPARATOR]);
 	for (row = 1; row < LINES - 3; row++)
-		mvaddch(row, tree_win_w, ACS_VLINE);
+		(void) mvaddch(row, tree_win_w, ACS_VLINE);
 }
 
 static void do_update_view(int full)
@@ -881,11 +945,13 @@ static void do_update_view(int full)
 			update_track_window();
 		editable_unlock();
 		draw_separator();
+		update_filterline();
 		break;
 	case SORTED_VIEW:
 		editable_lock();
 		update_sorted_window();
 		editable_unlock();
+		update_filterline();
 		break;
 	case PLAYLIST_VIEW:
 		editable_lock();
@@ -934,13 +1000,12 @@ static void do_update_statusline(void)
 	if (player_info.ti)
 		duration = player_info.ti->duration;
 
+	vol_left = vol_right = vol = -1;
 	if (soft_vol) {
 		vol_left = soft_vol_l;
 		vol_right = soft_vol_r;
 		vol = (vol_left + vol_right + 1) / 2;
-	} else if (!volume_max) {
-		vol_left = vol_right = vol = -1;
-	} else {
+	} else if (volume_max && volume_l >= 0 && volume_r >= 0) {
 		vol_left = scale_to_percentage(volume_l, volume_max);
 		vol_right = scale_to_percentage(volume_r, volume_max);
 		vol = (vol_left + vol_right + 1) / 2;
@@ -1317,6 +1382,20 @@ void update_statusline(void)
 	post_update();
 }
 
+void update_filterline(void)
+{
+	if (cur_view != TREE_VIEW && cur_view != SORTED_VIEW)
+		return;
+	if (lib_live_filter) {
+		char buf[512];
+		int w;
+		bkgdset(pairs[CURSED_STATUSLINE]);
+		snprintf(buf, sizeof(buf), "filtered: %s", lib_live_filter);
+		w = clamp(strlen(buf) + 2, COLS/4, COLS/2);
+		sprint(LINES-4, COLS-w, buf, w);
+	}
+}
+
 void info_msg(const char *format, ...)
 {
 	va_list ap;
@@ -1325,6 +1404,11 @@ void info_msg(const char *format, ...)
 	vsnprintf(error_buf, sizeof(error_buf), format, ap);
 	va_end(ap);
 
+	if (client_fd != -1) {
+		write_all(client_fd, error_buf, strlen(error_buf));
+		write_all(client_fd, "\n", 1);
+	}
+
 	msg_is_error = 0;
 
 	update_commandline();
@@ -1340,6 +1424,10 @@ void error_msg(const char *format, ...)
 	va_end(ap);
 
 	d_print("%s\n", error_buf);
+	if (client_fd != -1) {
+		write_all(client_fd, error_buf, strlen(error_buf));
+		write_all(client_fd, "\n", 1);
+	}
 
 	msg_is_error = 1;
 	error_count++;
@@ -1431,7 +1519,17 @@ void search_not_found(void)
 			break;
 		}
 	}
-	info_msg("%s not found: %s", what, search_str ? : "");
+	info_msg("%s not found: %s", what, search_str ? search_str : "");
+}
+
+void set_client_fd(int fd)
+{
+	client_fd = fd;
+}
+
+int get_client_fd(void)
+{
+	return client_fd;
 }
 
 void set_view(int view)
@@ -1538,24 +1636,24 @@ static void clear_error(void)
 
 static void spawn_status_program(void)
 {
-	static const char *status_strs[] = { "stopped", "playing", "paused" };
+	enum player_status status;
 	const char *stream_title = NULL;
 	char *argv[32];
-	int i, status;
+	int i;
 
 	if (status_display_program == NULL || status_display_program[0] == 0)
 		return;
 
 	player_info_lock();
 	status = player_info.status;
-	if (status == 1 && player_info.ti && is_url(player_info.ti->filename))
+	if (status == PLAYER_STATUS_PLAYING && player_info.ti && is_url(player_info.ti->filename))
 		stream_title = get_stream_title();
 
 	i = 0;
 	argv[i++] = xstrdup(status_display_program);
 
 	argv[i++] = xstrdup("status");
-	argv[i++] = xstrdup(status_strs[status]);
+	argv[i++] = xstrdup(player_status_names[status]);
 	if (player_info.ti) {
 		static const char *keys[] = {
 			"artist", "album", "discnumber", "tracknumber", "title", "date", NULL
@@ -1601,7 +1699,7 @@ static void spawn_status_program(void)
 	argv[i++] = NULL;
 	player_info_unlock();
 
-	if (spawn(argv, &status) == -1)
+	if (spawn(argv, NULL, 0) == -1)
 		error_msg("couldn't run `%s': %s", status_display_program, strerror(errno));
 	for (i = 0; argv[i]; i++)
 		free(argv[i]);
@@ -1784,6 +1882,20 @@ static void handle_ch(uchar ch)
 	}
 }
 
+static void handle_escape(int c)
+{
+	clear_error();
+	if (input_mode == NORMAL_MODE) {
+		normal_mode_ch(c + 128);
+	} else if (input_mode == COMMAND_MODE) {
+		command_mode_escape(c);
+		update_commandline();
+	} else if (input_mode == SEARCH_MODE) {
+		search_mode_escape(c);
+		update_commandline();
+	}
+}
+
 static void handle_key(int key)
 {
 	clear_error();
@@ -1814,6 +1926,15 @@ static void u_getch(void)
 		return;
 	}
 
+	/* escape sequence */
+	if (key == 0x1B) {
+		int e_key = getch();
+		if (e_key != ERR && e_key != 0) {
+			handle_escape(e_key);
+			return;
+		}
+	}
+
 	ch = (unsigned char)key;
 	while (bit > 0 && ch & mask) {
 		mask >>= 1;
@@ -1822,7 +1943,7 @@ static void u_getch(void)
 	if (bit == 7) {
 		/* ascii */
 		u = ch;
-	} else {
+	} else if (using_utf8) {
 		int count;
 
 		u = ch & ((1 << bit) - 1);
@@ -1836,7 +1957,8 @@ static void u_getch(void)
 			u = (u << 6) | (ch & 63);
 			count--;
 		}
-	}
+	} else
+		u = ch | U_INVALID_MASK;
 	handle_ch(u);
 }
 
@@ -1945,25 +2067,8 @@ static void main_loop(void)
 			item = next;
 		}
 
-		if (FD_ISSET(0, &set)) {
-			if (using_utf8) {
-				u_getch();
-			} else {
-				int key = getch();
-
-				if (key != ERR && key != 0) {
-					if (key > 255) {
-						handle_key(key);
-					} else {
-						uchar ch = key;
-
-						if (ch > 0x7f)
-							ch |= U_INVALID_MASK;
-						handle_ch(ch);
-					}
-				}
-			}
-		}
+		if (FD_ISSET(0, &set))
+			u_getch();
 	}
 }
 
@@ -2032,8 +2137,8 @@ static void init_curses(void)
 
 	noecho();
 	if (has_colors()) {
-		start_color();
 #if HAVE_USE_DEFAULT_COLORS
+		start_color();
 		use_default_colors();
 #endif
 	}
@@ -2045,7 +2150,7 @@ static void init_curses(void)
 	 */
 	update_colors();
 
-	ptr = print_buffer;
+	ptr = tcap_buffer;
 	t_ts = tgetstr("ts", &ptr);
 	t_fs = tgetstr("fs", &ptr);
 	d_print("ts: %d fs: %d\n", !!t_ts, !!t_fs);
@@ -2108,13 +2213,19 @@ static void init_all(void)
 	if (!soft_vol)
 		mixer_open();
 
+	if (resume_cmus)
+		resume_load();
+
 	lib_autosave_filename = xstrjoin(cmus_config_dir, "/lib.pl");
 	pl_autosave_filename = xstrjoin(cmus_config_dir, "/playlist.pl");
+	play_queue_autosave_filename = xstrjoin(cmus_config_dir, "/queue.pl");
 	pl_filename = xstrdup(pl_autosave_filename);
 	lib_filename = xstrdup(lib_autosave_filename);
 
-	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB);
+	if (resume_cmus)
+		cmus_add(play_queue_append, play_queue_autosave_filename, FILE_TYPE_PL, JOB_TYPE_QUEUE);
 	cmus_add(pl_add_track, pl_autosave_filename, FILE_TYPE_PL, JOB_TYPE_PL);
+	cmus_add(lib_add_track, lib_autosave_filename, FILE_TYPE_PL, JOB_TYPE_LIB);
 
 	if (error_count) {
 		char buf[16];
@@ -2134,10 +2245,14 @@ static void exit_all(void)
 {
 	endwin();
 
+	if (resume_cmus)
+		resume_exit();
 	options_exit();
 
 	server_exit();
 	cmus_exit();
+	if (resume_cmus)
+		cmus_save(play_queue_for_each, play_queue_autosave_filename);
 	cmus_save(lib_for_each, lib_autosave_filename);
 	cmus_save(pl_for_each, pl_autosave_filename);
 
@@ -2217,16 +2332,18 @@ int main(int argc, char *argv[])
 	}
 
 	setlocale(LC_CTYPE, "");
+	setlocale(LC_COLLATE, "");
+	charset = getenv("CMUS_CHARSET");
+	if (!charset || !charset[0]) {
 #ifdef CODESET
-	charset = nl_langinfo(CODESET);
+		charset = nl_langinfo(CODESET);
 #else
-	charset = "ISO-8859-1";
+		charset = "ISO-8859-1";
 #endif
-	if (strcmp(charset, "UTF-8") == 0) {
-		using_utf8 = 1;
-	} else {
-		using_utf8 = 0;
 	}
+	if (strcmp(charset, "UTF-8") == 0)
+		using_utf8 = 1;
+
 	misc_init();
 	if (server_address == NULL)
 		server_address = xstrjoin(cmus_config_dir, "/socket");
